# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Quantizer Algorithm Manager Interface."""

import enum
import functools
from immutabledict import immutabledict
from ai_edge_quantizer import algorithm_manager_api
from ai_edge_quantizer import default_policy
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.nonlinear_quantize import float_casting
from ai_edge_quantizer.algorithms.uniform_quantize import common_quantize
from ai_edge_quantizer.algorithms.uniform_quantize import dequantized_weight_recovery
from ai_edge_quantizer.algorithms.uniform_quantize import hadamard_rotation
from ai_edge_quantizer.algorithms.uniform_quantize import mse
from ai_edge_quantizer.algorithms.uniform_quantize import naive_min_max_quantize
from ai_edge_quantizer.algorithms.uniform_quantize import octav


# TODO: b/399775701 - Clean up this file.

_TFLOpName = qtyping.TFLOperationName

_alg_manager_instance = algorithm_manager_api.AlgorithmManagerApi()

# Expose instance functions.
get_quantization_func = _alg_manager_instance.get_quantization_func
get_supported_ops = _alg_manager_instance.get_supported_ops
get_init_qsv_func = _alg_manager_instance.get_init_qsv_func
register_op_quant_config_validation_func = (
    _alg_manager_instance.register_op_quant_config_validation_func
)
register_config_check_policy_func = (
    _alg_manager_instance.register_config_check_policy
)
register_quantized_op = _alg_manager_instance.register_quantized_op
is_op_registered = _alg_manager_instance.is_op_registered
is_algorithm_registered = _alg_manager_instance.is_algorithm_registered
check_op_quantization_config = (
    _alg_manager_instance.check_op_quantization_config
)


# Quantization algorithms.
class AlgorithmName(str, enum.Enum):
  NO_QUANTIZE = "no_quantize"
  MIN_MAX_UNIFORM_QUANT = naive_min_max_quantize.ALGORITHM_KEY
  FLOAT_CASTING = float_casting.ALGORITHM_KEY
  DEQUANTIZED_WEIGHT_RECOVERY = dequantized_weight_recovery.ALGORITHM_KEY
  OCTAV = octav.ALGORITHM_KEY
  HADAMARD_ROTATION = hadamard_rotation.CUSTOM_OP_ALGORITHM_KEY
  DECOMPOSED_HADAMARD_ROTATION = hadamard_rotation.DECOMPOSED_ALGORITHM_KEY
  MSE = mse.ALGORITHM_KEY


### MIN/MAX_UNIFORM_QUANT ###

# Register MIN_MAX_UNIFORM_QUANT algorithm.
register_op_quant_config_validation_func(
    AlgorithmName.MIN_MAX_UNIFORM_QUANT,
    common_quantize.check_op_quantization_config,
)

# Register a config check policy for MIN_MAX_UNIFORM_QUANT algorithm.
register_config_check_policy_func(
    AlgorithmName.MIN_MAX_UNIFORM_QUANT,
    default_policy.DEFAULT_CONFIG_CHECK_POLICY,
)

MIN_MAX_OP_NAME_MATERIALIZE_FUNC_DICT = {
    _TFLOpName.INPUT: common_quantize.materialize_input,
    _TFLOpName.OUTPUT: common_quantize.materialize_output,
    _TFLOpName.FULLY_CONNECTED: common_quantize.materialize_fc_conv,
    _TFLOpName.BATCH_MATMUL: common_quantize.materialize_batch_matmul,
    _TFLOpName.CONV_2D: common_quantize.materialize_fc_conv,
    _TFLOpName.DEPTHWISE_CONV_2D: common_quantize.materialize_fc_conv,
    _TFLOpName.CONV_2D_TRANSPOSE: common_quantize.materialize_conv2d_transpose,
    _TFLOpName.RESHAPE: common_quantize.materialize_reshape,
    _TFLOpName.AVERAGE_POOL_2D: common_quantize.materialize_average_pool_2d,
    _TFLOpName.EMBEDDING_LOOKUP: common_quantize.materialize_embedding_lookup,
    _TFLOpName.SOFTMAX: common_quantize.materialize_softmax_and_logistic,
    _TFLOpName.TANH: common_quantize.materialize_tanh,
    _TFLOpName.TRANSPOSE: common_quantize.materialize_transpose,
    _TFLOpName.GELU: common_quantize.materialize_gelu,
    _TFLOpName.ADD: common_quantize.materialize_add,
    _TFLOpName.SUB: common_quantize.materialize_sub,
    _TFLOpName.MUL: common_quantize.materialize_mul,
    _TFLOpName.MEAN: common_quantize.materialize_mean,
    _TFLOpName.RSQRT: common_quantize.materialize_rsqrt,
    _TFLOpName.CONCATENATION: common_quantize.materialize_concatenation,
    _TFLOpName.STRIDED_SLICE: common_quantize.materialize_strided_slice,
    _TFLOpName.SPLIT: common_quantize.materialize_split,
    _TFLOpName.LOGISTIC: common_quantize.materialize_softmax_and_logistic,
    _TFLOpName.SLICE: common_quantize.materialize_slice,
    _TFLOpName.SUM: common_quantize.materialize_sum,
    _TFLOpName.SELECT: common_quantize.materialize_select,
    _TFLOpName.SELECT_V2: common_quantize.materialize_select_v2,
    _TFLOpName.DYNAMIC_UPDATE_SLICE: (
        common_quantize.materialize_dynamic_update_slice
    ),
    _TFLOpName.STABLEHLO_COMPOSITE: common_quantize.materialize_composite,
    _TFLOpName.PAD: common_quantize.materialize_pad,
    _TFLOpName.SQUARED_DIFFERENCE: (
        common_quantize.materialize_squared_difference
    ),
    _TFLOpName.MAX_POOL_2D: common_quantize.materialize_max_pool_2d,
    _TFLOpName.RESIZE_BILINEAR: common_quantize.materialize_resize_bilinear,
    _TFLOpName.RESIZE_NEAREST_NEIGHBOR: (
        common_quantize.materialize_resize_nearest_neighbor
    ),
    _TFLOpName.GATHER_ND: common_quantize.materialize_gather_nd,
    _TFLOpName.PACK: common_quantize.materialize_pack,
    _TFLOpName.UNPACK: common_quantize.materialize_unpack,
    _TFLOpName.DIV: common_quantize.materialize_div,
    _TFLOpName.BROADCAST_TO: common_quantize.materialize_broadcast_to,
    _TFLOpName.SQRT: common_quantize.materialize_sqrt,
    _TFLOpName.GATHER: common_quantize.materialize_gather,
    _TFLOpName.HARD_SWISH: common_quantize.materialize_hard_swish,
    _TFLOpName.MAXIMUM: common_quantize.materialize_maximum,
    _TFLOpName.PADV2: common_quantize.materialize_padv2,
    _TFLOpName.REDUCE_MIN: common_quantize.materialize_reduce_min,
    _TFLOpName.EQUAL: common_quantize.materialize_equal,
    _TFLOpName.NOT_EQUAL: common_quantize.materialize_not_equal,
    _TFLOpName.MIRROR_PAD: common_quantize.materialize_mirror_pad,
    _TFLOpName.SPACE_TO_DEPTH: common_quantize.materialize_space_to_depth,
    _TFLOpName.RELU: common_quantize.materialize_relu,
}
for op_name, materialize_func in MIN_MAX_OP_NAME_MATERIALIZE_FUNC_DICT.items():
  register_quantized_op(
      AlgorithmName.MIN_MAX_UNIFORM_QUANT,
      op_name,
      naive_min_max_quantize.init_qsvs,
      calibration_func=naive_min_max_quantize.min_max_calibrate,
      # Most of the materialize op functions are common for all algorithms
      # except for the function to get scale and zero point, i.e.,
      # get_tensor_quant_params. So we use functools.partial here to pass in the
      # common utility function and thealgorithm-specific function.
      materialize_func=functools.partial(
          materialize_func,
          naive_min_max_quantize.get_tensor_quant_params,
      ),
  )

### FLOAT_CASTING ###
register_op_quant_config_validation_func(
    AlgorithmName.FLOAT_CASTING,
    float_casting.check_op_quantization_config,
)

# Register a config check policy for FLOAT_CASTING algorithm.
# TODO: b/353780772 - Replace an empty policy for FLOAT_CASTING algorithm.
register_config_check_policy_func(
    AlgorithmName.FLOAT_CASTING, qtyping.ConfigCheckPolicyDict()
)

for op_name, materialize_func in zip(
    (
        _TFLOpName.FULLY_CONNECTED,
        _TFLOpName.CONV_2D,
        _TFLOpName.DEPTHWISE_CONV_2D,
        _TFLOpName.CONV_2D_TRANSPOSE,
        _TFLOpName.EMBEDDING_LOOKUP,
    ),
    (
        float_casting.materialize_fc_conv,
        float_casting.materialize_fc_conv,
        float_casting.materialize_fc_conv,
        float_casting.materialize_conv2d_transpose,
        float_casting.materialize_embedding_lookup,
    ),
):
  register_quantized_op(
      AlgorithmName.FLOAT_CASTING,
      op_name,
      float_casting.init_qsvs,
      calibration_func=float_casting.calibrate,
      materialize_func=materialize_func,
  )

### DEQUANTIZED_WEIGHT_RECOVERY ###
register_op_quant_config_validation_func(
    AlgorithmName.DEQUANTIZED_WEIGHT_RECOVERY,
    common_quantize.check_op_quantization_config,
)

register_config_check_policy_func(
    AlgorithmName.DEQUANTIZED_WEIGHT_RECOVERY,
    default_policy.DEFAULT_CONFIG_CHECK_POLICY,
)

DEQUANTIZED_WEIGHT_RECOVERY_OP_NAME_MATERIALIZE_FUNC_DICT = {
    _TFLOpName.FULLY_CONNECTED: common_quantize.materialize_fc_conv,
    _TFLOpName.CONV_2D: common_quantize.materialize_fc_conv,
    _TFLOpName.EMBEDDING_LOOKUP: common_quantize.materialize_embedding_lookup,
}

for (
    op_name,
    materialize_func,
) in DEQUANTIZED_WEIGHT_RECOVERY_OP_NAME_MATERIALIZE_FUNC_DICT.items():
  register_quantized_op(
      algorithm_key=AlgorithmName.DEQUANTIZED_WEIGHT_RECOVERY,
      tfl_op_name=op_name,
      init_qsv_func=dequantized_weight_recovery.init_qsvs,
      calibration_func=dequantized_weight_recovery.calibrate,
      # Most of the materialize op functions are common for all algorithms
      # except for the function to get scale and zero point, i.e.,
      # get_tensor_quant_params. So we use functools.partial here to pass in the
      # common utility function and the algorithm-specific function.
      materialize_func=functools.partial(
          materialize_func,
          dequantized_weight_recovery.get_tensor_quant_params,
      ),
  )


# Register OCTAV algorithm.
register_op_quant_config_validation_func(
    AlgorithmName.OCTAV,
    common_quantize.check_op_quantization_config,
)

# Register a config check policy for OCTAV algorithm.
register_config_check_policy_func(
    AlgorithmName.OCTAV,
    default_policy.DEFAULT_CONFIG_CHECK_POLICY,
)

_OCTAV_OP_NAME_MATERIALIZE_FUNC_DICT = immutabledict({
    _TFLOpName.INPUT: common_quantize.materialize_input,
    _TFLOpName.OUTPUT: common_quantize.materialize_output,
    _TFLOpName.FULLY_CONNECTED: common_quantize.materialize_fc_conv,
    _TFLOpName.BATCH_MATMUL: common_quantize.materialize_batch_matmul,
    _TFLOpName.CONV_2D: common_quantize.materialize_fc_conv,
    _TFLOpName.DEPTHWISE_CONV_2D: common_quantize.materialize_fc_conv,
    _TFLOpName.CONV_2D_TRANSPOSE: common_quantize.materialize_conv2d_transpose,
    _TFLOpName.RESHAPE: common_quantize.materialize_reshape,
    _TFLOpName.AVERAGE_POOL_2D: common_quantize.materialize_average_pool_2d,
    _TFLOpName.EMBEDDING_LOOKUP: common_quantize.materialize_embedding_lookup,
    _TFLOpName.SOFTMAX: common_quantize.materialize_softmax_and_logistic,
    _TFLOpName.TANH: common_quantize.materialize_tanh,
    _TFLOpName.TRANSPOSE: common_quantize.materialize_transpose,
    _TFLOpName.GELU: common_quantize.materialize_gelu,
    _TFLOpName.ADD: common_quantize.materialize_add,
    _TFLOpName.SUB: common_quantize.materialize_sub,
    _TFLOpName.MUL: common_quantize.materialize_mul,
    _TFLOpName.MEAN: common_quantize.materialize_mean,
    _TFLOpName.RSQRT: common_quantize.materialize_rsqrt,
    _TFLOpName.CONCATENATION: common_quantize.materialize_concatenation,
    _TFLOpName.STRIDED_SLICE: common_quantize.materialize_strided_slice,
    _TFLOpName.SPLIT: common_quantize.materialize_split,
    _TFLOpName.LOGISTIC: common_quantize.materialize_softmax_and_logistic,
    _TFLOpName.SLICE: common_quantize.materialize_slice,
    _TFLOpName.SUM: common_quantize.materialize_sum,
    _TFLOpName.SELECT: common_quantize.materialize_select,
    _TFLOpName.SELECT_V2: common_quantize.materialize_select_v2,
    _TFLOpName.DYNAMIC_UPDATE_SLICE: (
        common_quantize.materialize_dynamic_update_slice
    ),
    _TFLOpName.STABLEHLO_COMPOSITE: common_quantize.materialize_composite,
    _TFLOpName.PAD: common_quantize.materialize_pad,
    _TFLOpName.SQUARED_DIFFERENCE: (
        common_quantize.materialize_squared_difference
    ),
    _TFLOpName.MAX_POOL_2D: common_quantize.materialize_max_pool_2d,
    _TFLOpName.RESIZE_BILINEAR: common_quantize.materialize_resize_bilinear,
    _TFLOpName.RESIZE_NEAREST_NEIGHBOR: (
        common_quantize.materialize_resize_nearest_neighbor
    ),
    _TFLOpName.GATHER_ND: common_quantize.materialize_gather_nd,
    _TFLOpName.PACK: common_quantize.materialize_pack,
    _TFLOpName.UNPACK: common_quantize.materialize_unpack,
    _TFLOpName.DIV: common_quantize.materialize_div,
    _TFLOpName.BROADCAST_TO: common_quantize.materialize_broadcast_to,
    _TFLOpName.SQRT: common_quantize.materialize_sqrt,
    _TFLOpName.GATHER: common_quantize.materialize_gather,
    _TFLOpName.HARD_SWISH: common_quantize.materialize_hard_swish,
    _TFLOpName.MAXIMUM: common_quantize.materialize_maximum,
    _TFLOpName.PADV2: common_quantize.materialize_padv2,
    _TFLOpName.REDUCE_MIN: common_quantize.materialize_reduce_min,
    _TFLOpName.EQUAL: common_quantize.materialize_equal,
    _TFLOpName.NOT_EQUAL: common_quantize.materialize_not_equal,
    _TFLOpName.MIRROR_PAD: common_quantize.materialize_mirror_pad,
    _TFLOpName.SPACE_TO_DEPTH: common_quantize.materialize_space_to_depth,
    _TFLOpName.RELU: common_quantize.materialize_relu,
})

for op_name, materialize_func in _OCTAV_OP_NAME_MATERIALIZE_FUNC_DICT.items():
  register_quantized_op(
      AlgorithmName.OCTAV,
      op_name,
      naive_min_max_quantize.init_qsvs,
      calibration_func=naive_min_max_quantize.min_max_calibrate,
      materialize_func=functools.partial(
          materialize_func,
          octav.get_tensor_quant_params,
      ),
  )

# Register the Hadamard Rotation algorithm.
register_op_quant_config_validation_func(
    AlgorithmName.HADAMARD_ROTATION,
    common_quantize.check_op_quantization_config,
)

# Register a config check policy for the Hadamard Rotation algorithm.
register_config_check_policy_func(
    AlgorithmName.HADAMARD_ROTATION,
    default_policy.DEFAULT_CONFIG_CHECK_POLICY,
)

# Register specialized hadamard rotation materialize functions.
_HADAMARD_ROTATION_OP_NAME_MATERIALIZE_FUNC_DICT = immutabledict({
    _TFLOpName.FULLY_CONNECTED: (
        hadamard_rotation.materialize_fully_connected_custom_op
    ),
    _TFLOpName.EMBEDDING_LOOKUP: (
        hadamard_rotation.materialize_embedding_lookup_custom_op
    ),
})
for (
    op_name,
    materialize_func,
) in _HADAMARD_ROTATION_OP_NAME_MATERIALIZE_FUNC_DICT.items():
  register_quantized_op(
      AlgorithmName.HADAMARD_ROTATION,
      op_name,
      naive_min_max_quantize.init_qsvs,
      calibration_func=naive_min_max_quantize.min_max_calibrate,
      materialize_func=materialize_func,
  )

register_op_quant_config_validation_func(
    AlgorithmName.DECOMPOSED_HADAMARD_ROTATION,
    common_quantize.check_op_quantization_config,
)

register_config_check_policy_func(
    AlgorithmName.DECOMPOSED_HADAMARD_ROTATION,
    default_policy.DEFAULT_CONFIG_CHECK_POLICY,
)

_DECOMPOSED_HADAMARD_ROTATION_OP_NAME_MATERIALIZE_FUNC_DICT = immutabledict({
    _TFLOpName.FULLY_CONNECTED: (
        hadamard_rotation.materialize_fully_connected_decomposed
    ),
    _TFLOpName.EMBEDDING_LOOKUP: (
        hadamard_rotation.materialize_embedding_lookup_decomposed
    ),
})
for (
    op_name,
    materialize_func,
) in _DECOMPOSED_HADAMARD_ROTATION_OP_NAME_MATERIALIZE_FUNC_DICT.items():
  register_quantized_op(
      AlgorithmName.DECOMPOSED_HADAMARD_ROTATION,
      op_name,
      naive_min_max_quantize.init_qsvs,
      calibration_func=naive_min_max_quantize.min_max_calibrate,
      materialize_func=materialize_func,
  )


# Register the MSE algorithm.
register_op_quant_config_validation_func(
    AlgorithmName.MSE,
    common_quantize.check_op_quantization_config,
)

# Register a config check policy for the MSE algorithm.
register_config_check_policy_func(
    AlgorithmName.MSE,
    default_policy.DEFAULT_CONFIG_CHECK_POLICY,
)

# Register specialized MSE materialize functions.
_MSE_OP_NAME_MATERIALIZE_FUNC_DICT = immutabledict({
    _TFLOpName.FULLY_CONNECTED: common_quantize.materialize_fc_conv,
    _TFLOpName.EMBEDDING_LOOKUP: common_quantize.materialize_embedding_lookup,
    _TFLOpName.CONV_2D: common_quantize.materialize_fc_conv,
    _TFLOpName.DEPTHWISE_CONV_2D: common_quantize.materialize_fc_conv,
    _TFLOpName.CONV_2D_TRANSPOSE: common_quantize.materialize_conv2d_transpose,
})
for (
    op_name,
    materialize_func,
) in _MSE_OP_NAME_MATERIALIZE_FUNC_DICT.items():
  register_quantized_op(
      AlgorithmName.MSE,
      op_name,
      naive_min_max_quantize.init_qsvs,
      calibration_func=naive_min_max_quantize.min_max_calibrate,
      materialize_func=functools.partial(
          materialize_func,
          mse.get_tensor_quant_params,
      ),
  )



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""function for validating output models."""

from collections.abc import Callable, Iterable
import dataclasses
import json
import math
import pathlib
from typing import Any, Optional, Union

import numpy as np

from ai_edge_quantizer.utils import tfl_interpreter_utils as utils
import os
import io


_DEFAULT_SIGNATURE_KEY = utils.DEFAULT_SIGNATURE_KEY


@dataclasses.dataclass(frozen=True)
class SingleSignatureComparisonResult:
  """Comparison result for a single signature.

  Attributes:
    error_metric: The name of the error metric used for comparison.
    input_tensors: A dictionary of input tensor name and its value.
    output_tensors: A dictionary of output tensor name and its value.
    constant_tensors: A dictionary of constant tensor name and its value.
    intermediate_tensors: A dictionary of intermediate tensor name and its
      value.
  """

  error_metric: str
  input_tensors: dict[str, Any]
  output_tensors: dict[str, Any]
  constant_tensors: dict[str, Any]
  intermediate_tensors: dict[str, Any]


class ComparisonResult:
  """Comparison result for a model.

  Attributes:
    comparison_results: A dictionary of signature key and its comparison result.
  """

  def __init__(self, reference_model: bytes, target_model: bytes):
    """Initialize the ComparisonResult object.

    Args:
      reference_model: Model which will be used as the reference.
      target_model: Target model which will be compared against the reference.
        We expect target_model and reference_model to have the same graph
        structure.
    """
    self._reference_model = reference_model
    self._target_model = target_model
    self._comparison_results: dict[str, SingleSignatureComparisonResult] = {}

  def get_signature_comparison_result(
      self, signature_key: str = _DEFAULT_SIGNATURE_KEY
  ) -> SingleSignatureComparisonResult:
    """Get the comparison result for a signature.

    Args:
      signature_key: The signature key to be used for invoking the models.

    Returns:
      A SingleSignatureComparisonResult object.
    """
    if signature_key not in self._comparison_results:
      raise ValueError(
          f'{signature_key} is not in the comparison_results. Available'
          f' signature keys are: {self.available_signature_keys()}'
      )
    return self._comparison_results[signature_key]

  def available_signature_keys(self) -> list[str]:
    """Get the available signature keys in the comparison result."""
    return list(self._comparison_results.keys())

  def add_new_signature_results(
      self,
      error_metric: str,
      comparison_result: dict[str, float],
      signature_key: str = _DEFAULT_SIGNATURE_KEY,
      validate_output_tensors_only: bool = False,
  ) -> None:
    """Add a new signature result to the comparison result.

    Args:
      error_metric: The name of the error metric used for comparison.
      comparison_result: A dictionary of tensor name and its value.
      signature_key: The model signature that the comparison_result belongs to.
      validate_output_tensors_only: If True, only compare output tensors.
        Otherwise, compare all tensors.

    Raises:
      ValueError: If the signature_key is already in the comparison_results.
    """
    if signature_key in self._comparison_results:
      raise ValueError(f'{signature_key} is already in the comparison_results.')

    result = {key: float(value) for key, value in comparison_result.items()}

    output_tensor_results = {}
    for name in utils.get_output_tensor_names(
        self._reference_model, signature_key
    ):
      output_tensor_results[name] = result.pop(name)

    input_tensor_results = {}
    constant_tensor_results = {}
    if validate_output_tensors_only:
      # No intermediate tensors are validated when validate_output_tensors_only
      # is True.
      result = {}
    else:
      for name in utils.get_input_tensor_names(
          self._reference_model, signature_key
      ):
        if name in result:
          input_tensor_results[name] = result.pop(name)

      # Only get constant tensors from the main subgraph of the signature.
      subgraph_index = utils.get_signature_main_subgraph_index(
          utils.create_tfl_interpreter(self._reference_model),
          signature_key,
      )
      for name in utils.get_constant_tensor_names(
          self._reference_model,
          subgraph_index,
      ):
        if name in result:
          constant_tensor_results[name] = result.pop(name)

    self._comparison_results[signature_key] = SingleSignatureComparisonResult(
        error_metric=error_metric,
        input_tensors=input_tensor_results,
        output_tensors=output_tensor_results,
        constant_tensors=constant_tensor_results,
        intermediate_tensors=result,
    )

  def get_all_tensor_results(self) -> dict[str, Any]:
    """Get all the tensor results in a single dictionary.

    Returns:
      A dictionary of tensor name and its value.
    """
    result = {}
    for _, signature_comparison_result in self._comparison_results.items():
      result.update(signature_comparison_result.input_tensors)
      result.update(signature_comparison_result.output_tensors)
      result.update(signature_comparison_result.constant_tensors)
      result.update(signature_comparison_result.intermediate_tensors)
    return result

  def get_model_size_reduction(self) -> tuple[int, float]:
    """Get the model size reduction in bytes and percentage."""
    reduced_model_size = len(self._reference_model) - len(self._target_model)
    reduction_perc = reduced_model_size / len(self._reference_model) * 100
    return reduced_model_size, reduction_perc

  def save(self, save_folder: str, model_name: str) -> None:
    """Saves the model comparison result.

    Args:
      save_folder: Path to the folder to save the comparison result.
      model_name: Name of the model.

    Raises:
      RuntimeError: If no quantized model is available.
    """
    reduced_model_size, reduction_ratio = self.get_model_size_reduction()
    result = {
        'reduced_size_bytes': reduced_model_size,
        'reduced_size_percentage': reduction_ratio,
    }
    for signature, comparison_result in self._comparison_results.items():
      result[str(signature)] = {
          'error_metric': comparison_result.error_metric,
          'input_tensors': comparison_result.input_tensors,
          'output_tensors': comparison_result.output_tensors,
          'constant_tensors': comparison_result.constant_tensors,
          'intermediate_tensors': comparison_result.intermediate_tensors,
      }
    result_save_path = str(
        pathlib.Path(save_folder) / (model_name + '_comparison_result.json')
    )
    with open(result_save_path, 'w') as output_file_handle:
      output_file_handle.write(json.dumps(result))

    # TODO: b/365578554 - Remove after ME is updated to use the new json format.
    color_threshold = [0.05, 0.1, 0.2, 0.4, 1, 10, 100]
    json_object = create_json_for_model_explorer(
        self,
        threshold=color_threshold,
    )
    json_save_path = str(
        pathlib.Path(save_folder)
        / (model_name + '_comparison_result_me_input.json')
    )
    with open(json_save_path, 'w') as output_file_handle:
      output_file_handle.write(json_object)


def _setup_validation_interpreter(
    model: bytes,
    signature_input: dict[str, Any],
    signature_key: Optional[str],
    use_xnnpack: bool,
    num_threads: int,
    preserve_all_tensors: bool = True,
) -> tuple[Any, int, dict[str, Any]]:
  """Setup the interpreter for validation given a signature key.

  Args:
    model: The model to be validated.
    signature_input: A dictionary of input tensor name and its value.
    signature_key: The signature key to be used for invoking the models. If the
      model only has one signature, this can be set to None.
    use_xnnpack: Whether to use xnnpack for the interpreter.
    num_threads: The number of threads to use for the interpreter.
    preserve_all_tensors: Whether to preserve all tensors.

  Returns:
    A tuple of interpreter, subgraph_index and tensor_name_to_details.
  """

  interpreter = utils.create_tfl_interpreter(
      tflite_model=model,
      use_xnnpack=use_xnnpack,
      num_threads=num_threads,
      preserve_all_tensors=preserve_all_tensors,
  )
  utils.invoke_interpreter_signature(
      interpreter, signature_input, signature_key
  )
  # Only validate tensors from the main subgraph of the signature.
  subgraph_index = utils.get_signature_main_subgraph_index(
      interpreter, signature_key
  )
  tensor_name_to_details = utils.get_tensor_name_to_details_map(
      interpreter,
      subgraph_index,
  )
  return interpreter, subgraph_index, tensor_name_to_details


# TODO: b/330797129 - Enable multi-threaded evaluation.
def compare_model(
    reference_model: bytes,
    target_model: bytes,
    test_data: dict[str, Iterable[dict[str, Any]]],
    error_metric: str,
    compare_fn: Callable[[Any, Any], float],
    use_xnnpack: bool = True,
    num_threads: int = 16,
    validate_output_tensors_only: bool = False,
) -> ComparisonResult:
  """Compares model tensors over a model signature using the compare_fn.

  This function will run the model signature on the provided dataset over and
  compare all the tensors (cached) using the compare_fn (e.g., mean square
  error).

  Args:
    reference_model: Model which will be used as the reference
    target_model: Target model which will be compared against the reference. We
      expect reference_model and target_model have the inputs and outputs
      signature.
    test_data: A dictionary of signature key and its correspending test input
      data that will be used for comparison.
    error_metric: The name of the error metric used for comparison.
    compare_fn: a comparison function to be used for calculating the statistics,
      this function must be taking in two ArrayLike strcuture and output a
      single float value.
    use_xnnpack: Whether to use xnnpack for the interpreter.
    num_threads: The number of threads to use for the interpreter.
    validate_output_tensors_only: If True, only compare output tensors.
      Otherwise, compare all tensors.

  Returns:
    A ComparisonResult object.
  """
  preserve_all_tensors = not validate_output_tensors_only
  model_comparion_result = ComparisonResult(reference_model, target_model)
  for signature_key, signature_inputs in test_data.items():
    comparison_results = {}
    for signature_input in signature_inputs:
      # Invoke the signature on both interpreters.
      ref_interpreter, ref_subgraph_index, ref_tensor_name_to_details = (
          _setup_validation_interpreter(
              reference_model,
              signature_input,
              signature_key,
              use_xnnpack,
              num_threads,
              preserve_all_tensors=preserve_all_tensors,
          )
      )
      targ_interpreter, targ_subgraph_index, targ_tensor_name_to_details = (
          _setup_validation_interpreter(
              target_model,
              signature_input,
              signature_key,
              use_xnnpack,
              num_threads,
              preserve_all_tensors=preserve_all_tensors,
          )
      )
      # Compare the cached tensor value
      tensor_names_to_compare = (
          utils.get_output_tensor_names(reference_model, signature_key)
          if validate_output_tensors_only
          else list(ref_tensor_name_to_details.keys())
      )

      for tensor_name in tensor_names_to_compare:
        detail = ref_tensor_name_to_details[tensor_name]
        if detail['dtype'] == np.object_:
          continue
        # Ignore tensors where any dimension of the shape is 0.
        if not np.all(detail['shape']):
          continue
        if tensor_name in targ_tensor_name_to_details:
          if tensor_name not in comparison_results:
            comparison_results[tensor_name] = []

          reference_data = utils.get_tensor_data(
              ref_interpreter, detail, ref_subgraph_index
          )
          target_data = utils.get_tensor_data(
              targ_interpreter,
              targ_tensor_name_to_details[tensor_name],
              targ_subgraph_index,
          )
          comparison_results[tensor_name].append(
              compare_fn(target_data, reference_data)
          )

    agregated_results = {}
    for tensor_name in comparison_results:
      agregated_results[tensor_name] = np.mean(comparison_results[tensor_name])
    model_comparion_result.add_new_signature_results(
        error_metric,
        agregated_results,
        signature_key,
        validate_output_tensors_only,
    )
  return model_comparion_result


def create_json_for_model_explorer(
    data: ComparisonResult, threshold: list[Union[int, float]]
) -> str:
  """create a dict type that can be exported as json for model_explorer to use.

  Args:
    data: Output from compare_model function
    threshold: A list of numbers representing thresholds for model_exlorer to
      display different colors

  Returns:
    A string represents the json format accepted by model_explorer
  """
  data = data.get_all_tensor_results()
  color_scheme = []
  results = {name: {'value': float(value)} for name, value in data.items()}
  if threshold:
    green = 255
    gradient = math.floor(255 / len(threshold))
    for val in threshold:
      color_scheme.append({'value': val, 'bgColor': f'rgb(200, {green}, 0)'})
      green = max(0, green - gradient)

  return json.dumps({
      'results': results,
      'thresholds': color_scheme,
  })



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Tests for tfl_flatbuffer_utils.py."""

import pathlib
import numpy as np
import absl.testing.absltest as absltest
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils


TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile("../tests/models")


# TODO: b/328830092 - Add test cases for model require buffer offset.
class FlatbufferUtilsTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self._test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH) / "conv_fc_mnist.tflite"
    )

    self._test_model = tfl_flatbuffer_utils.read_model(self._test_model_path)

  def test_get_model_buffer(self):
    model_buffer = tfl_flatbuffer_utils.get_model_buffer(self._test_model_path)
    file_stats = pathlib.Path(self._test_model_path).stat()
    self.assertLen(model_buffer, file_stats.st_size)

  def test_parse_op_tensors(self):
    subgraph0 = self._test_model.subgraphs[0]
    conv2d_op = subgraph0.operators[0]
    op_tensors = tfl_flatbuffer_utils.parse_op_tensors(
        conv2d_op, subgraph0.tensors
    )
    # conv2d have three inputs and one output
    self.assertLen(op_tensors, 4)

    average_pool_op = subgraph0.operators[1]
    op_tensors = tfl_flatbuffer_utils.parse_op_tensors(
        average_pool_op, subgraph0.tensors
    )
    # averagepool have one input and one output
    self.assertLen(op_tensors, 2)

  def test_parse_fc_bmm_conv_tensors(self):
    subgraph0 = self._test_model.subgraphs[0]
    conv2d_op = subgraph0.operators[0]
    inputs, weight, bias, output = (
        tfl_flatbuffer_utils.parse_fc_bmm_conv_tensors(
            conv2d_op, subgraph0.tensors
        )
    )
    self.assertEqual(tuple(inputs.shape), (1, 28, 28, 1))
    self.assertEqual(tuple(weight.shape), (8, 3, 3, 1))
    self.assertEqual(tuple(bias.shape), (8,))
    self.assertEqual(tuple(output.shape), (1, 28, 28, 8))

    fc_with_bias = subgraph0.operators[3]
    inputs, weight, bias, output = (
        tfl_flatbuffer_utils.parse_fc_bmm_conv_tensors(
            fc_with_bias,
            subgraph0.tensors,
        )
    )
    self.assertEqual(tuple(inputs.shape), (1, 1568))
    self.assertEqual(tuple(weight.shape), (32, 1568))
    self.assertEqual(tuple(bias.shape), (32,))
    self.assertEqual(tuple(output.shape), (1, 32))

    fc_no_bias = subgraph0.operators[4]
    inputs, weight, bias, output = (
        tfl_flatbuffer_utils.parse_fc_bmm_conv_tensors(
            fc_no_bias,
            subgraph0.tensors,
        )
    )
    self.assertEqual(tuple(inputs.shape), (1, 32))
    self.assertEqual(tuple(weight.shape), (10, 32))
    self.assertIsNone(bias)
    self.assertEqual(tuple(output.shape), (1, 10))

  def test_buffer_to_tensors(self):
    buffer_to_tensor_map = tfl_flatbuffer_utils.buffer_to_tensors(
        self._test_model
    )
    # Read from Netron/Model Explorer
    tensors = buffer_to_tensor_map[6]
    self.assertLen(tensors, 1)
    conv2d_filter_tensor = tensors[0]
    self.assertEqual(tuple(conv2d_filter_tensor.shape), (8, 3, 3, 1))

  def test_buffer_to_tensors_has_unique_values(self):
    test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / "constant_tensor_and_buffer_only_sharing_weight_fcs.tflite",
    )
    test_model = tfl_flatbuffer_utils.read_model(test_model_path)
    buffer_to_tensor_map = tfl_flatbuffer_utils.buffer_to_tensors(test_model)
    self.assertLen(buffer_to_tensor_map, 7)
    # The following buffer is shared by two tensors, each shared by two FC ops.
    # This is where before we had multiple enrties for the same tensor.
    self.assertLen(buffer_to_tensor_map[2], 2)
    got_tensor_names = [
        tfl_flatbuffer_utils.get_tensor_name(tensor)
        for tensor in buffer_to_tensor_map[2]
    ]
    self.assertEqual(
        got_tensor_names,
        ["arith.constant", "arith.constant1"],
    )

  def test_get_tensor_name(self):
    subgraph0 = self._test_model.subgraphs[0]
    subgraph_tensors = subgraph0.tensors
    conv2d_op = subgraph0.operators[0]
    weight_tensor = subgraph_tensors[conv2d_op.inputs[1]]
    weight_tensor_name = tfl_flatbuffer_utils.get_tensor_name(weight_tensor)
    self.assertEqual(weight_tensor_name, "sequential/conv2d/Conv2D")

  # TODO: b/325123193 - test tensor with data outside of flatbuffer.
  def test_get_tensor_data(self):
    subgraph0 = self._test_model.subgraphs[0]
    subgraph_tensors = subgraph0.tensors
    conv2d_op = subgraph0.operators[0]
    # Check tensor with data
    weight_tensor = subgraph_tensors[conv2d_op.inputs[1]]
    weight_tensor_data = tfl_flatbuffer_utils.get_tensor_data(
        weight_tensor, self._test_model.buffers
    )
    self.assertEqual(
        tuple(weight_tensor.shape), tuple(weight_tensor_data.shape)  # pytype: disable=attribute-error
    )
    self.assertAlmostEqual(weight_tensor_data[0][0][0][0], -0.12941549718379974)

    # Check tensor with no data
    input_tensor = subgraph_tensors[conv2d_op.inputs[0]]
    input_tensor_data = tfl_flatbuffer_utils.get_tensor_data(
        input_tensor, self._test_model.buffers
    )
    self.assertIsNone(input_tensor_data)

  def test_has_same_quantization_succeeds(self):
    tensor0, tensor1 = self._test_model.subgraphs[0].tensors[:2]
    tensor0.quantization.scale = np.array([1, 2, 3]).astype(np.float32)
    tensor0.quantization.zeroPoint = np.array([3, 2, 1]).astype(np.int32)
    tensor1.quantization.scale = np.array([1, 2, 3]).astype(np.float32)
    tensor1.quantization.zeroPoint = np.array([3, 2, 1]).astype(np.int32)
    self.assertTrue(
        tfl_flatbuffer_utils.has_same_quantization(tensor0, tensor1)
    )

  def test_has_same_quantization_succeds_not_quantized(self):
    tensor0, tensor1 = self._test_model.subgraphs[0].tensors[:2]
    tensor0.type = 10
    self.assertTrue(
        tfl_flatbuffer_utils.has_same_quantization(tensor0, tensor1)
    )

  def test_has_same_quantization_fails_different_scale(self):
    tensor0, tensor1 = self._test_model.subgraphs[0].tensors[:2]
    tensor1.quantization.scale = np.array([1, 2, 3]).astype(np.float32)
    self.assertFalse(
        tfl_flatbuffer_utils.has_same_quantization(tensor0, tensor1)
    )

  def test_has_same_quantization_fails_different_zp(self):
    tensor0, tensor1 = self._test_model.subgraphs[0].tensors[:2]
    tensor0.quantization.scale = np.array([1, 2, 3]).astype(np.float32)
    tensor0.quantization.zeroPoint = np.array([3, 2, 1]).astype(np.int32)
    tensor1.quantization.scale = np.array([1, 2, 3]).astype(np.float32)
    tensor1.quantization.zeroPoint = np.array([1, 2, 3]).astype(np.int32)
    self.assertFalse(
        tfl_flatbuffer_utils.has_same_quantization(tensor0, tensor1)
    )

  def test_check_is_float_model_true_when_model_is_float(self):
    test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH) / "conv_fc_mnist.tflite"
    )
    model = tfl_flatbuffer_utils.read_model(test_model_path)
    self.assertTrue(tfl_flatbuffer_utils.is_float_model(model))

  def test_check_is_float_model_false_when_model_is_quantized(self):
    test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH) / "mnist_quantized.tflite"
    )
    model = tfl_flatbuffer_utils.read_model(test_model_path)
    self.assertFalse(tfl_flatbuffer_utils.is_float_model(model))

  def test_get_subgraph_input_output_operators(self):
    subgraph = self._test_model.subgraphs[0]
    input_op, output_op = (
        tfl_flatbuffer_utils.get_subgraph_input_output_operators(subgraph)
    )
    self.assertEqual(input_op.op_key, qtyping.TFLOperationName.INPUT)
    self.assertEmpty(input_op.inputs)
    self.assertListEqual(list(input_op.outputs), [0])
    self.assertEqual(output_op.op_key, qtyping.TFLOperationName.OUTPUT)
    self.assertListEqual(list(output_op.inputs), [12])
    self.assertEmpty(output_op.outputs)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""E2E tests for the quantizer using a toy MNIST model."""

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from ai_edge_litert.tools import flatbuffer_utils
from ai_edge_quantizer import qtyping
from ai_edge_quantizer import quantizer
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_interpreter_utils


_ComputePrecision = qtyping.ComputePrecision
_OpName = qtyping.TFLOperationName
_TensorQuantConfig = qtyping.TensorQuantizationConfig
_OpQuantConfig = qtyping.OpQuantizationConfig

_RNG = np.random.default_rng(66)


def _get_dummy_data(num_samples):
  data = []
  for _ in range(num_samples):
    data.append(
        {'conv2d_input': _RNG.uniform(size=(1, 28, 28, 1)).astype(np.float32)}
    )
  return data


def _get_calibration_data(num_samples: int = 256):
  samples = _get_dummy_data(num_samples)
  data = {
      tfl_interpreter_utils.DEFAULT_SIGNATURE_KEY: samples,
  }
  return data


def _get_test_data(num_samples: int = 8):
  return _get_calibration_data(num_samples)


class MNISTTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.float_model_path = test_utils.get_path_to_datafile(
        'models/conv_fc_mnist.tflite'
    )
    self._quantizer = quantizer.Quantizer(self.float_model_path)

  def _assert_all_tensors_integer(self, quantized_model):
    """Asserts that all tensors in the given model are integers."""
    quantized_model = flatbuffer_utils.read_model_from_bytearray(
        quantized_model
    )
    for graph in quantized_model.subgraphs:
      for op in graph.operators:
        if op.inputs is None:
          break
        for input_idx in op.inputs:
          tensor = graph.tensors[input_idx]
          tensor_type = flatbuffer_utils.type_to_name(tensor.type)
          op_name = flatbuffer_utils.opcode_to_name(
              quantized_model, op.opcodeIndex
          )
          if 'QUANTIZE' in op_name:
            continue  # Skip quantize ops as they may have float inputs.

          self.assertFalse(tensor_type.startswith('FLOAT'))

  @parameterized.product(
      test_case=[
          # Tuple holds compute precision and whether to use explicit
          # dequantize.
          (_ComputePrecision.FLOAT, True),  # WEIGHT_ONLY.
          (_ComputePrecision.INTEGER, False),  # DRQ.
      ],
      symmetric_weight=[True, False],
      granularity=[
          qtyping.QuantGranularity.CHANNELWISE,
          qtyping.QuantGranularity.TENSORWISE,
      ],
  )
  def test_mnist_toy_model_int8_weight_only(
      self, test_case, symmetric_weight, granularity
  ):
    compute_precision, explicit_dequantize = test_case
    # asym DRQ is not supported
    # TODO: b/335254997 - fail when trying to use unsupported recipe.
    if compute_precision == _ComputePrecision.INTEGER and not symmetric_weight:
      return
    self._quantizer.update_quantization_recipe(
        regex='.*',
        operation_name=_OpName.FULLY_CONNECTED,
        op_config=_OpQuantConfig(
            weight_tensor_config=_TensorQuantConfig(
                num_bits=8,
                symmetric=symmetric_weight,
                granularity=granularity,
            ),
            compute_precision=compute_precision,
            explicit_dequantize=explicit_dequantize,
        ),
    )
    _ = self._quantizer.quantize()
    # Check model size.
    self.assertLess(len(self._quantizer._result.quantized_model), 55000)

    comparison_result = self._quantizer.validate(error_metrics='mse')
    self._check_comparison_result(
        comparison_result,
        weight_tolerance=1e-2
        if granularity == qtyping.QuantGranularity.CHANNELWISE
        else 1e-1,
        logits_tolerance=1e-2
        if granularity == qtyping.QuantGranularity.CHANNELWISE
        else 1e-1,
        output_tolerance=1e-4
        if granularity == qtyping.QuantGranularity.CHANNELWISE
        else 1e-2,
    )

  @parameterized.product(
      test_case=[
          # Tuple holds compute precision and whether to use explicit
          # dequantize.
          (_ComputePrecision.FLOAT, True),  # WEIGHT_ONLY.
          (_ComputePrecision.INTEGER, False),  # DRQ.
      ],
      symmetric_weight=[True, False],
  )
  def test_mnist_toy_model_int4_weight_only(self, test_case, symmetric_weight):

    compute_precision, explicit_dequantize = test_case
    # Asym DRQ is not supported.
    # TODO: b/335254997 - Fail when trying to use unsupported recipe.
    if compute_precision == _ComputePrecision.INTEGER and not symmetric_weight:
      return
    self._quantizer.update_quantization_recipe(
        regex='.*',
        operation_name=_OpName.FULLY_CONNECTED,
        op_config=_OpQuantConfig(
            weight_tensor_config=_TensorQuantConfig(
                num_bits=4,
                symmetric=symmetric_weight,
                granularity=qtyping.QuantGranularity.CHANNELWISE,
            ),
            compute_precision=compute_precision,
            explicit_dequantize=explicit_dequantize,
        ),
    )
    _ = self._quantizer.quantize()
    # Check model size.
    self.assertLess(len(self._quantizer._result.quantized_model), 30000)

    comparison_result = self._quantizer.validate(error_metrics='mse')
    self._check_comparison_result(
        comparison_result,
        weight_tolerance=1e-3,
        logits_tolerance=2,
        output_tolerance=1e-2,
    )

  def test_mnist_toy_model_fp16_weight_only(self):
    self._quantizer.update_quantization_recipe(
        regex='.*',
        algorithm_key=quantizer.AlgorithmName.FLOAT_CASTING,
        operation_name=_OpName.FULLY_CONNECTED,
        op_config=_OpQuantConfig(
            weight_tensor_config=_TensorQuantConfig(
                num_bits=16, dtype=qtyping.TensorDataType.FLOAT
            ),
            compute_precision=_ComputePrecision.FLOAT,
        ),
    )
    _ = self._quantizer.quantize()
    # Check model size.
    self.assertLess(len(self._quantizer._result.quantized_model), 105000)

    comparison_result = self._quantizer.validate(error_metrics='mse')
    self._check_comparison_result(
        comparison_result,
        weight_tolerance=1e-5,
        logits_tolerance=1e-5,
        output_tolerance=1e-5,
    )

  @parameterized.parameters(
      '../recipes/default_a8w8_recipe.json',
      '../recipes/default_a16w8_recipe.json',
  )
  def test_mnist_toy_model_full_integer(self, recipe_path):
    recipe_path = test_utils.get_path_to_datafile(recipe_path)
    self._quantizer.load_quantization_recipe(recipe_path)
    self.assertTrue(self._quantizer.need_calibration)
    calibration_result = self._quantizer.calibrate(_get_calibration_data())
    quant_result = self._quantizer.quantize(calibration_result)
    # Check model size.
    self.assertLess(len(quant_result.quantized_model), 55000)

    comparison_result = self._quantizer.validate(
        error_metrics='mse', test_data=_get_test_data()
    )
    self._check_comparison_result(
        comparison_result,
        weight_tolerance=1e-2,
        logits_tolerance=1e-1,
        output_tolerance=1e-4,
    )

  # TODO: b/345503484 - Check weight tensor type of the quantized model.
  def _check_comparison_result(
      self,
      comparison_result,
      weight_tolerance,
      logits_tolerance,
      output_tolerance,
  ):
    # TODO: b/357959309 - Use comparison result directly for testing.
    comparison_result = comparison_result.get_all_tensor_results()
    # Check weight tensors.
    conv_weight_mse = comparison_result['sequential/conv2d/Conv2D']
    self.assertLess(conv_weight_mse, weight_tolerance)
    fc1_weight_mse = comparison_result['arith.constant1']
    self.assertLess(fc1_weight_mse, weight_tolerance)
    fc2_weight_mse = comparison_result['arith.constant']
    self.assertLess(fc2_weight_mse, weight_tolerance)
    # check logits.
    logits_mse = comparison_result['sequential/dense_1/MatMul']
    self.assertLess(logits_mse, logits_tolerance)
    # check final output.
    output_mse = comparison_result['StatefulPartitionedCall:0']
    self.assertLess(output_mse, output_tolerance)

  @parameterized.parameters(
      '../recipes/default_a8w8_recipe.json',
      '../recipes/default_a16w8_recipe.json',
  )
  def test_mnist_toy_model_full_integer_requantize_success(self, recipe_path):
    partly_quantized_model_path = test_utils.get_path_to_datafile(
        '../tests/models/partly_quantized_mnist.tflite'
    )
    recipe_path = test_utils.get_path_to_datafile(recipe_path)
    qt = quantizer.Quantizer(partly_quantized_model_path, recipe_path)
    self.assertTrue(qt.need_calibration)
    calibration_result = qt.calibrate(_get_calibration_data())
    quant_result = qt.quantize(calibration_result)

    # Check if the output model has `integer-only` tensors.
    self._assert_all_tensors_integer(quant_result.quantized_model)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Default quantization policy."""

import collections
import copy
import json
from typing import Any, Union
from ai_edge_litert.tools import flatbuffer_utils
from ai_edge_quantizer import qtyping
from ai_edge_litert import schema_py_generated as schema  # pylint:disable=g-direct-tensorflow-import

_TFLOpName = qtyping.TFLOperationName
_OpQuantizationConfig = qtyping.OpQuantizationConfig
_TensorQuantizationConfig = qtyping.TensorQuantizationConfig
_ComputePrecision = qtyping.ComputePrecision
_Granularity = qtyping.QuantGranularity
_INT = qtyping.TensorDataType.INT

# Default config check policy in JSON format. It has two keys: "configs" and
# "ops_per_config". "configs" is a dictionary mapping from config name to config
# content. "ops_per_config" is a dictionary mapping from config name to a list
# of op names.
DEFAULT_JSON_POLICY = """
{
  "configs": {
    "dynamic_wi8_afp32": {
      "weight_tensor_config": {
        "num_bits": 8,
        "symmetric": [true],
        "granularity": ["CHANNELWISE", "TENSORWISE"],
        "dtype": "INT"
      },
      "explicit_dequantize": false,
      "compute_precision": "INTEGER"
    },
    "dynamic_wi4_afp32": {
      "weight_tensor_config": {
        "num_bits": 4,
        "symmetric": [true],
        "granularity": ["CHANNELWISE", "TENSORWISE"],
        "dtype": "INT"
      },
      "explicit_dequantize": false,
      "compute_precision": "INTEGER"
    },
    "dynamic_wi4_afp32_blockwise": {
      "weight_tensor_config": {
        "num_bits": 4,
        "symmetric": [true],
        "granularity": ["BLOCKWISE_32", "BLOCKWISE_64", "BLOCKWISE_128", "BLOCKWISE_256"],
        "dtype": "INT"
      },
      "explicit_dequantize": false,
      "compute_precision": "INTEGER"
    },
    "static_wi8_ai16": {
      "activation_tensor_config": {
        "num_bits": 16,
        "symmetric": [true],
        "granularity": ["TENSORWISE"],
        "dtype": "INT"
      },
      "weight_tensor_config": {
        "num_bits": 8,
        "symmetric": [true],
        "granularity": ["CHANNELWISE", "TENSORWISE"],
        "dtype": "INT"
      },
      "explicit_dequantize": false,
      "compute_precision": "INTEGER"
    },
    "static_wi4_ai16": {
      "activation_tensor_config": {
        "num_bits": 16,
        "symmetric": [true],
        "granularity": ["TENSORWISE"],
        "dtype": "INT"
      },
      "weight_tensor_config": {
        "num_bits": 4,
        "symmetric": [true],
        "granularity": ["CHANNELWISE", "TENSORWISE"],
        "dtype": "INT"
      },
      "explicit_dequantize": false,
      "compute_precision": "INTEGER"
    },
    "static_wi8_ai8": {
      "activation_tensor_config": {
        "num_bits": 8,
        "symmetric": [true, false],
        "granularity": ["TENSORWISE"],
        "dtype": "INT"
      },
      "weight_tensor_config": {
        "num_bits": 8,
        "symmetric": [true],
        "granularity": ["CHANNELWISE", "TENSORWISE"],
        "dtype": "INT"
      },
      "explicit_dequantize": false,
      "compute_precision": "INTEGER"
    },
    "static_wi4_ai8": {
      "activation_tensor_config": {
        "num_bits": 8,
        "symmetric": [true, false],
        "granularity": ["TENSORWISE"],
        "dtype": "INT"
      },
      "weight_tensor_config": {
        "num_bits": 4,
        "symmetric": [true],
        "granularity": ["CHANNELWISE", "TENSORWISE"],
        "dtype": "INT"
      },
      "explicit_dequantize": false,
      "compute_precision": "INTEGER"
    },
    "weightonly_wi8_afp32": {
      "weight_tensor_config": {
        "num_bits": 8,
        "symmetric": [true, false],
        "granularity": ["CHANNELWISE", "TENSORWISE"],
        "dtype": "INT"
      },
      "explicit_dequantize": true,
      "compute_precision": "FLOAT"
    },
    "weightonly_wi4_afp32": {
      "weight_tensor_config": {
        "num_bits": 4,
        "symmetric": [true, false],
        "granularity": ["CHANNELWISE", "TENSORWISE"],
        "dtype": "INT"
      },
      "explicit_dequantize": true,
      "compute_precision": "FLOAT"
    }
  },
  "ops_per_config": {
    "static_wi8_ai16": [
      "ADD",
      "AVERAGE_POOL_2D",
      "BATCH_MATMUL",
      "CONCATENATION",
      "CONV_2D",
      "CONV_2D_TRANSPOSE",
      "DEPTHWISE_CONV_2D",
      "FULLY_CONNECTED",
      "GELU",
      "LOGISTIC",
      "MEAN",
      "MUL",
      "RESHAPE",
      "RSQRT",
      "SOFTMAX",
      "SPLIT",
      "STRIDED_SLICE",
      "SUB",
      "TANH",
      "TRANSPOSE",
      "INPUT",
      "OUTPUT",
      "SLICE",
      "SUM",
      "SELECT",
      "SELECT_V2",
      "DYNAMIC_UPDATE_SLICE",
      "SELECT_V2",
      "STABLEHLO_COMPOSITE",
      "PAD",
      "MAX_POOL_2D",
      "RESIZE_BILINEAR",
      "RESIZE_NEAREST_NEIGHBOR",
      "GATHER_ND",
      "PACK",
      "UNPACK",
      "DIV",
      "BROADCAST_TO",
      "SQRT",
      "GATHER",
      "MAXIMUM",
      "PADV2",
      "REDUCE_MIN",
      "EQUAL",
      "NOT_EQUAL",
      "MIRROR_PAD",
      "RELU"
    ],
    "static_wi8_ai8": [
      "ADD",
      "AVERAGE_POOL_2D",
      "BATCH_MATMUL",
      "CONCATENATION",
      "FULLY_CONNECTED",
      "CONV_2D",
      "CONV_2D_TRANSPOSE",
      "DEPTHWISE_CONV_2D",
      "GELU",
      "LOGISTIC",
      "MEAN",
      "MUL",
      "RESHAPE",
      "RSQRT",
      "SOFTMAX",
      "SPLIT",
      "STRIDED_SLICE",
      "SUB",
      "TANH",
      "TRANSPOSE",
      "INPUT",
      "OUTPUT",
      "SLICE",
      "SUM",
      "SELECT",
      "SELECT_V2",
      "DYNAMIC_UPDATE_SLICE",
      "SELECT_V2",
      "STABLEHLO_COMPOSITE",
      "PAD",
      "SQUARED_DIFFERENCE",
      "MAX_POOL_2D",
      "RESIZE_BILINEAR",
      "RESIZE_NEAREST_NEIGHBOR",
      "GATHER_ND",
      "PACK",
      "UNPACK",
      "DIV",
      "BROADCAST_TO",
      "SQRT",
      "GATHER",
      "HARD_SWISH",
      "MAXIMUM",
      "PADV2",
      "REDUCE_MIN",
      "EQUAL",
      "NOT_EQUAL",
      "MIRROR_PAD",
      "SPACE_TO_DEPTH",
      "RELU"
    ],
    "static_wi4_ai8": ["FULLY_CONNECTED", "CONV_2D", "INPUT", "OUTPUT"],
    "static_wi4_ai16": ["FULLY_CONNECTED", "CONV_2D", "INPUT", "OUTPUT"],
    "dynamic_wi8_afp32": [
      "BATCH_MATMUL",
      "CONV_2D",
      "CONV_2D_TRANSPOSE",
      "DEPTHWISE_CONV_2D",
      "EMBEDDING_LOOKUP",
      "FULLY_CONNECTED"
    ],
    "dynamic_wi4_afp32": ["FULLY_CONNECTED", "EMBEDDING_LOOKUP", "CONV_2D"],
    "dynamic_wi4_afp32_blockwise": ["EMBEDDING_LOOKUP", "FULLY_CONNECTED"],
    "weightonly_wi8_afp32": [
      "BATCH_MATMUL",
      "CONV_2D",
      "CONV_2D_TRANSPOSE",
      "DEPTHWISE_CONV_2D",
      "EMBEDDING_LOOKUP",
      "FULLY_CONNECTED"
    ],
    "weightonly_wi4_afp32": ["BATCH_MATMUL", "FULLY_CONNECTED", "EMBEDDING_LOOKUP", "CONV_2D"]
  }
}
"""
QUANTIZABLE_COMPOSITES = [
    "od" + "ml.npu_call",
    "od" + "ml.rms_norm",
    "od" + "ml.l2_norm",
]


def _unroll_json_config(
    json_config: dict[str, Any],
) -> list[_OpQuantizationConfig]:
  """Unrolls the config.

  Args:
    json_config: JSON config to be unrolled.

  Returns:
    quant_configs: A list of unrolled configs.
  """

  # Unroll activation configs first.
  activation_configs = []
  if "activation_tensor_config" in json_config:
    for symmetric in json_config["activation_tensor_config"]["symmetric"]:
      for granularity in json_config["activation_tensor_config"]["granularity"]:
        tensor_config = {
            "num_bits": json_config["activation_tensor_config"]["num_bits"],
            "symmetric": symmetric,
            "granularity": granularity,
            "dtype": json_config["activation_tensor_config"]["dtype"],
        }
        activation_configs.append(
            qtyping.TensorQuantizationConfig.from_dict(tensor_config)
        )

  # Then unroll weight configs and turn them into quantization configs.
  quant_configs = []
  weight_configs = []
  for symmetric in json_config["weight_tensor_config"]["symmetric"]:
    for granularity in json_config["weight_tensor_config"]["granularity"]:
      tensor_config = {
          "num_bits": json_config["weight_tensor_config"]["num_bits"],
          "symmetric": symmetric,
          "granularity": granularity,
          "dtype": json_config["weight_tensor_config"]["dtype"],
      }
      weight_configs.append(
          qtyping.TensorQuantizationConfig.from_dict(tensor_config)
      )

      if activation_configs:
        for activation_config in activation_configs:
          quant_configs.append(
              qtyping.OpQuantizationConfig(
                  activation_tensor_config=activation_config,
                  weight_tensor_config=qtyping.TensorQuantizationConfig.from_dict(
                      tensor_config
                  ),
                  compute_precision=json_config["compute_precision"],
                  explicit_dequantize=json_config["explicit_dequantize"],
              )
          )
      else:
        for weight_config in weight_configs:
          quant_configs.append(
              qtyping.OpQuantizationConfig(
                  weight_tensor_config=weight_config,
                  compute_precision=json_config["compute_precision"],
                  explicit_dequantize=json_config["explicit_dequantize"],
              )
          )

  return quant_configs


# TODO: b/401024954 - Have a better way to specify recipes based on op options.
def is_non_quantizable_composite_op(
    op: Union[schema.Operator, schema.OperatorT],
) -> bool:
  """Checks if the operator is a non-quantizable composite op.

  We may want to quantize an op only when its has certain options.
  Policies/recipes
  are not aware of op options, so it is checked here.

  Args:
    op: The operator to check.

  Returns:
    True if the operator is conditionally unquantized, False otherwise.
  """
  if opts := flatbuffer_utils.get_options_as(
      op, schema.StableHLOCompositeOptionsT
  ):
    name = opts.name.decode("utf-8")
    if name not in QUANTIZABLE_COMPOSITES:
      return True

  return False


def update_default_config_policy(raw_json_policy: str):
  """Updates the default config check policy."""
  json_policy_content = json.loads(raw_json_policy)

  # Unroll the json config and add the configs to the policy.
  policy = collections.OrderedDict()
  for json_policy_config in json_policy_content["ops_per_config"]:
    unrolled_configs = _unroll_json_config(
        json_policy_content["configs"][json_policy_config]
    )

    for op in json_policy_content["ops_per_config"][json_policy_config]:
      op_name = _TFLOpName(op)
      quant_configs = copy.deepcopy(unrolled_configs)
      if op in policy.keys():
        quant_configs += policy[op_name]
      policy[op_name] = quant_configs

  return policy


DEFAULT_CONFIG_CHECK_POLICY = update_default_config_policy(DEFAULT_JSON_POLICY)



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Quantization recipe module."""

from ai_edge_quantizer import algorithm_manager
from ai_edge_quantizer import qtyping
from ai_edge_quantizer import recipe_manager

AlgorithmName = algorithm_manager.AlgorithmName


def dynamic_wi8_afp32(
    algorithm_key: AlgorithmName = AlgorithmName.MIN_MAX_UNIFORM_QUANT,
) -> recipe_manager.ModelQuantizationRecipe:
  """Returns a dynamic quantization recipe with int8 weights and float32 activation.

  All supported ops will be quantized with int8 weights and float32 activations,
  which will be dynamically quantized to int8 during inference to enable int8
  compute. The model quality may suffer due to the on-the-fly quantization. If
  quality is a concern, consider using weight-only quantization.

  Args:
    algorithm_key: The algorithm to use for quantization.

  Returns:
    A dynamic quantization recipe.
  """
  rp_manager = recipe_manager.RecipeManager()
  rp_manager.add_dynamic_config(
      regex='.*',
      operation_name=qtyping.TFLOperationName.ALL_SUPPORTED,
      num_bits=8,
      algorithm_key=algorithm_key,
  )
  return rp_manager.get_quantization_recipe()


def dynamic_wi4_afp32(
    algorithm_key: AlgorithmName = AlgorithmName.MIN_MAX_UNIFORM_QUANT,
) -> recipe_manager.ModelQuantizationRecipe:
  """Returns a dynamic quantization recipe with int4 weights and float32 activation.

  All supported ops will be quantized with int4 weights and float32 activations,
  which will be dynamically quantized to int4 during inference to enable int4
  compute.

  Args:
    algorithm_key: The algorithm to use for quantization.

  Returns:
    A dynamic quantization recipe.
  """
  rp_manager = recipe_manager.RecipeManager()
  rp_manager.add_dynamic_config(
      regex='.*',
      operation_name=qtyping.TFLOperationName.ALL_SUPPORTED,
      num_bits=4,
      algorithm_key=algorithm_key,
  )
  return rp_manager.get_quantization_recipe()


def weight_only_wi8_afp32(
    algorithm_key: AlgorithmName = AlgorithmName.MIN_MAX_UNIFORM_QUANT,
) -> recipe_manager.ModelQuantizationRecipe:
  """Returns a weight-only quantization recipe with int8 weights and float32 activation.

  All supported ops will be quantized with int8 weights and float32 activations.
  The weights will be explicitly dequantized before being fed into the op to
  enable float compute thus retain model quality. If latency is a concern,
  consider using dynamic range quantization.

  Args:
    algorithm_key: The algorithm to use for quantization.

  Returns:
    A weight-only quantization recipe.
  """
  rp_manager = recipe_manager.RecipeManager()
  rp_manager.add_weight_only_config(
      regex='.*',
      operation_name=qtyping.TFLOperationName.ALL_SUPPORTED,
      num_bits=8,
      algorithm_key=algorithm_key,
  )
  return rp_manager.get_quantization_recipe()


def weight_only_wi4_afp32(
    algorithm_key: AlgorithmName = AlgorithmName.MIN_MAX_UNIFORM_QUANT,
) -> recipe_manager.ModelQuantizationRecipe:
  """Returns a weight-only quantization recipe with int4 weights and float32 activation.

  All supported ops will be quantized with int4 weights and float32 activations.
  The weights will be explicitly dequantized before being fed into the op to
  enable float compute thus retain model quality.

  Args:
    algorithm_key: The algorithm to use for quantization.

  Returns:
    A weight-only quantization recipe.
  """
  rp_manager = recipe_manager.RecipeManager()
  rp_manager.add_weight_only_config(
      regex='.*',
      operation_name=qtyping.TFLOperationName.ALL_SUPPORTED,
      num_bits=4,
      algorithm_key=algorithm_key,
  )
  return rp_manager.get_quantization_recipe()


def static_wi8_ai8(
    algorithm_key: AlgorithmName = AlgorithmName.MIN_MAX_UNIFORM_QUANT,
) -> recipe_manager.ModelQuantizationRecipe:
  """Returns a static quantization recipe with int8 weights and int8 activations.

  All supported ops will be quantized with int8 weights and int8 activations.
  Calibration is needed to use this recipe.

  Args:
    algorithm_key: The algorithm to use for quantization.

  Returns:
    A static quantization recipe.
  """
  rp_manager = recipe_manager.RecipeManager()
  rp_manager.add_static_config(
      regex='.*',
      operation_name=qtyping.TFLOperationName.ALL_SUPPORTED,
      activation_num_bits=8,
      weight_num_bits=8,
      algorithm_key=algorithm_key,
  )
  return rp_manager.get_quantization_recipe()


def static_wi8_ai16(
    algorithm_key: AlgorithmName = AlgorithmName.MIN_MAX_UNIFORM_QUANT,
) -> recipe_manager.ModelQuantizationRecipe:
  """Returns a static quantization recipe with int8 weights and int16 activations.

  All supported ops will be quantized with int8 weights and int16 activations.
  Calibration is needed to use this recipe.

  Args:
    algorithm_key: The algorithm to use for quantization.

  Returns:
    A static quantization recipe.
  """
  rp_manager = recipe_manager.RecipeManager()
  rp_manager.add_static_config(
      regex='.*',
      operation_name=qtyping.TFLOperationName.ALL_SUPPORTED,
      activation_num_bits=16,
      weight_num_bits=8,
      algorithm_key=algorithm_key,
  )
  return rp_manager.get_quantization_recipe()


def dynamic_legacy_wi8_afp32():
  """Returns a dynamic quantization legacy recipe with int8 weights and float32 activation.

  The difference between this and dynamic_wi8_afp32 is that this recipe sets
  min_weight_elements to 1024 to match the old quantizer behavior.
  """
  return [
      dict({
          'regex': '.*',
          'operation': '*',
          'algorithm_key': 'min_max_uniform_quantize',
          'op_config': {
              'weight_tensor_config': {
                  'num_bits': 8,
                  'symmetric': True,
                  'granularity': 'CHANNELWISE',
                  'dtype': 'INT',
                  'block_size': 0,
              },
              'compute_precision': 'INTEGER',
              'explicit_dequantize': False,
              'skip_checks': False,
              'min_weight_elements': 1024,
          },
      })
  ]
